@@ -10,7 +10,7 @@ is_pypy = platform.python_implementation() == "PyPy"
 from .tasks import create_task, TaskFactory
 import sys
 import logging
-
+import uuid
 
 @ffi.callback("void(uws_res_t*, const char*, size_t, bool, void*)")
 def wsgi_on_data_handler(res, chunk, chunk_length, is_end, user_data):
@@ -25,6 +25,85 @@ def wsgi_on_data_handler(res, chunk, chunk_length, is_end, user_data):
             data_response._ptr,
         )
 
+@ffi.callback("void(uws_res_t*, void*)")
+def wsgi_on_data_ref_abort_handler(res, user_data):
+    data_retry = ffi.from_handle(user_data)
+    data_retry.aborted = True
+    if data_retry.id is not None:
+        data_retry.app._data_refs.pop(data_retry.id, None)    
+    
+
+@ffi.callback("bool(uws_res_t*, uintmax_t, void*)")
+def wsgi_on_writable_handler(res, offset, user_data):
+    data_retry = ffi.from_handle(user_data)
+    if data_retry.aborted:
+        return False
+    
+    chunks = data_retry.chunks
+    last_sended_offset = data_retry.last_offset
+    ssl = data_retry.app.server.SSL
+    content_length = data_retry.content_length
+    
+    data = chunks[0]
+    data_size = len(data)
+    last_offset = int(lib.uws_res_get_write_offset(ssl, res))
+    if last_sended_offset != last_offset:
+        offset = last_offset - last_sended_offset
+        data = data[offset:data_size]
+        data_size = len(data)
+        if data_size == 0:
+            chunks.pop(0)
+        
+        if len(chunks) == 0:
+            logging.error(AssertionError("Content-Length do not match sended content"))
+            lib.uws_res_close(
+                ssl,
+                res
+            )
+            if data_retry.id is not None:
+                data_retry.app._data_refs.pop(data_retry.id, None)
+    
+            return True
+        data = chunks[0]
+
+    result = lib.uws_res_try_end(
+        ssl,
+        res,
+        data,
+        data_size,
+        content_length,
+        0,
+    )
+    has_responded = bool(result.has_responded)
+    ok = bool(result.ok)
+    data_retry.last_offset = int(lib.uws_res_get_write_offset(ssl, res))
+
+    if ok:
+        chunks.pop(0)
+        if not has_responded and len(chunks) == 0:
+            logging.error(AssertionError("Content-Length do not match sended content"))
+            lib.uws_res_close(
+                ssl,
+                res
+            )
+            if data_retry.id is not None:
+                data_retry.app._data_refs.pop(data_retry.id, None)
+        elif has_responded and data_retry.id is not None:
+            data_retry.app._data_refs.pop(data_retry.id, None)
+    elif not has_responded and len(chunks) == 0:
+        logging.error(AssertionError("Content-Length do not match sended content"))
+        lib.uws_res_close(
+            ssl,
+            res
+        )
+        if data_retry.id is not None:
+            data_retry.app._data_refs.pop(data_retry.id, None)
+    elif has_responded and data_retry.id is not None:
+        data_retry.app._data_refs.pop(data_retry.id, None)
+    
+
+    return ok
+
 
 class WSGIBody:
     def __init__(self, buffer):
@@ -115,14 +194,25 @@ class WSGIBody:
 
 
 class WSGIDataResponse:
-    def __init__(self, app, environ, start_response, aborted, buffer, on_data):
+    def __init__(self, app, environ, start_response, buffer, on_data):
         self.buffer = buffer
-        self.aborted = aborted
         self._ptr = ffi.new_handle(self)
         self.on_data = on_data
         self.environ = environ
         self.app = app
         self.start_response = start_response
+        self.id = None
+        self.aborted = False
+
+class WSGIRetryDataSend:
+    def __init__(self, app, chunks, content_length, last_offset):
+        self.chunks = chunks
+        self._ptr = ffi.new_handle(self)
+        self.app = app
+        self.content_length = content_length
+        self.last_offset = last_offset
+        self.id = None
+        self.aborted = False
 
 
 @ffi.callback("void(uws_res_t*, void*)")
@@ -131,9 +221,11 @@ def wsgi_corked_response_start_handler(res, user_data):
     data_response.on_data(data_response, res)
 
 
-@ffi.callback("void(int, uws_res_t*, socketify_asgi_data request, void*, bool*)")
-def wsgi(ssl, response, info, user_data, aborted):
+
+@ffi.callback("void(int, uws_res_t*, socketify_asgi_data request, void*)")
+def wsgi(ssl, response, info, user_data):
     app = ffi.from_handle(user_data)
+
     # reusing the dict is slower than cloning because we need to clear HTTP headers
     environ = dict(app.BASIC_ENVIRON)
 
@@ -163,9 +255,10 @@ def wsgi(ssl, response, info, user_data, aborted):
     headers_set = None
     headers_written = False
     status_text = None
-
+    is_chunked = False
+    content_length = -1
     def write_headers(headers):
-        nonlocal headers_written, headers_set, status_text
+        nonlocal headers_written, headers_set, status_text, content_length, is_chunked
         if headers_written or not headers_set:
             return
 
@@ -183,20 +276,33 @@ def wsgi(ssl, response, info, user_data, aborted):
                 if (
                     key == "content-length"
                     or key == "Content-Length"
-                    or key == "Transfer-Encoding"
+                ):
+                    content_length = int(value)
+                    continue  # auto generated by try_end
+                if (
+                    key == "Transfer-Encoding"
                     or key == "transfer-encoding"
                 ):
-                    continue  # auto
+                    is_chunked = str(value) == "chunked"
+                    if is_chunked:
+                        continue
+                
                 key_data = key.encode("utf-8")
             elif isinstance(key, bytes):
                 # this is faster than using .lower()
                 if (
                     key == b"content-length"
                     or key == b"Content-Length"
-                    or key == b"Transfer-Encoding"
+                ):
+                    content_length = int(value)
+                    continue  # auto
+                if (
+                    key == b"Transfer-Encoding"
                     or key == b"transfer-encoding"
                 ):
-                    continue  # auto
+                    is_chunked = str(value) == "chunked"
+                    if is_chunked:
+                        continue
                 key_data = key
 
             if isinstance(value, str):
@@ -216,6 +322,10 @@ def wsgi(ssl, response, info, user_data, aborted):
             lib.uws_res_write_header(
                 ssl, response, key_data, len(key_data), value_data, len(value_data)
             )
+        # no content-length
+        if content_length < 0:
+            is_chunked = True
+        content_length = ffi.cast("uintmax_t", content_length)
 
     def start_response(status, headers, exc_info=None):
         nonlocal headers_set, status_text
@@ -233,9 +343,12 @@ def wsgi(ssl, response, info, user_data, aborted):
         status_text = status
 
         def write(data):
+            nonlocal is_chunked
             if not headers_written:
                 write_headers(headers_set)
-
+            # will allow older frameworks only with is_chunked
+            is_chunked = True 
+                
             if isinstance(data, bytes):
                 lib.uws_res_write(ssl, response, data, len(data))
             elif isinstance(data, str):
@@ -244,32 +357,69 @@ def wsgi(ssl, response, info, user_data, aborted):
 
         return write
 
+    
+    
+    
     # check for body
     if bool(info.has_content):
         WSGI_INPUT = BytesIO()
         environ["wsgi.input"] = WSGIBody(WSGI_INPUT)
-
         def on_data(data_response, response):
-            if bool(data_response.aborted[0]):
+            last_offset = -1
+            data_retry = None
+            failed_chunks = None
+
+
+            if data_response.aborted:
                 return
 
             ssl = data_response.app.server.SSL
             data_response.environ["CONTENT_LENGTH"] = str(
                 data_response.buffer.getbuffer().nbytes
             )
+            if data_response.id is not None:
+                data_response.app._data_refs.pop(data_response.id, None)
+                
             app_iter = data_response.app.wsgi(
                 data_response.environ, data_response.start_response
             )
+            
             try:
                 for data in app_iter:
-                    if data and not headers_written:
-                        write_headers(headers_set)
-
-                    if isinstance(data, bytes):
-                        lib.uws_res_write(ssl, response, data, len(data))
-                    elif isinstance(data, str):
-                        data = data.encode("utf-8")
-                        lib.uws_res_write(ssl, response, data, len(data))
+                    if data:
+                        if not headers_written:
+                            write_headers(headers_set)
+                            
+                        if is_chunked:
+                            if isinstance(data, bytes):
+                                lib.uws_res_write(ssl, response, data, len(data))
+                            elif isinstance(data, str):
+                                data = data.encode("utf-8")
+                                lib.uws_res_write(ssl, response, data, len(data))
+                        else:
+                            if isinstance(data, str):
+                                data = data.encode("utf-8")
+                            if failed_chunks:
+                                failed_chunks.append(data)
+                            else:
+                                last_offset = int(lib.uws_res_get_write_offset(ssl, response))
+                                result = lib.uws_res_try_end(
+                                    ssl,
+                                    response,
+                                    data,
+                                    len(data),
+                                    content_length,
+                                    0,
+                                )
+                                # this should be very very rare for HTTP
+                                if not bool(result.ok):
+                                    failed_chunks = []
+                                    # just mark the chunks
+                                    failed_chunks.append(data)
+                                    # add on writable handler
+                                    data_retry = WSGIRetryDataSend(
+                                        app, failed_chunks, content_length, last_offset
+                                    )
 
             except Exception as error:
                 logging.exception(error)
@@ -278,27 +428,72 @@ def wsgi(ssl, response, info, user_data, aborted):
                     app_iter.close()
 
             if not headers_written:
-                write_headers(headers_set)
-            lib.uws_res_end_without_body(ssl, response, 0)
-
+                write_headers(headers_set)         
+            if is_chunked:
+                lib.uws_res_end_without_body(ssl, response, 0)
+            elif data_retry is not None:
+                _id = uuid.uuid4()
+                app._data_refs[_id] = data_retry
+                lib.uws_res_on_aborted(ssl, response, wsgi_on_data_ref_abort_handler, data_retry._ptr)
+                lib.uws_res_on_writable(ssl, response, wsgi_on_writable_handler, data_retry._ptr)
+            elif result is None or (not bool(result.has_responded) and bool(result.ok)): # not reachs Content-Length
+                logging.error(AssertionError("Content-Length do not match sended content"))
+                lib.uws_res_close(
+                    ssl,
+                    response
+                )
         data_response = WSGIDataResponse(
-            app, environ, start_response, aborted, WSGI_INPUT, on_data
+            app, environ, start_response, WSGI_INPUT, on_data
         )
-
+        _id = uuid.uuid4()
+        data_response.id = _id
+        app._data_refs[_id] = data_response
+        lib.uws_res_on_aborted(ssl, response, wsgi_on_data_ref_abort_handler, data_response._ptr)
         lib.uws_res_on_data(ssl, response, wsgi_on_data_handler, data_response._ptr)
     else:
+        failed_chunks = None
+        last_offset = -1
+        data_retry = None
         environ["wsgi.input"] = None
         app_iter = app.wsgi(environ, start_response)
+        result = None
         try:
             for data in app_iter:
-                if data and not headers_written:
-                    write_headers(headers_set)
+                if data:
+                    if not headers_written:
+                        write_headers(headers_set)
+                    if is_chunked:
+                        if isinstance(data, bytes):
+                            lib.uws_res_write(ssl, response, data, len(data))
+                        elif isinstance(data, str):
+                            data = data.encode("utf-8")
+                            lib.uws_res_write(ssl, response, data, len(data))
+                    else:
+                        if isinstance(data, str):
+                            data = data.encode("utf-8")
+                        if failed_chunks: # if failed once, will fail again later
+                            failed_chunks.append(data)
+                        else:
+                            last_offset = int(lib.uws_res_get_write_offset(ssl, response))
+                            result = lib.uws_res_try_end(
+                                ssl,
+                                response,
+                                data,
+                                len(data),
+                                content_length,
+                                0,
+                            )
+                            # this should be very very rare fot HTTP
+                            if not bool(result.ok):
+                                failed_chunks = []
+                                # just mark the chunks
+                                failed_chunks.append(data)
+                                # add on writable handler
+                                data_retry = WSGIRetryDataSend(
+                                    app, failed_chunks, content_length, last_offset
+                                )
+                        
 
-                if isinstance(data, bytes):
-                    lib.uws_res_write(ssl, response, data, len(data))
-                elif isinstance(data, str):
-                    data = data.encode("utf-8")
-                    lib.uws_res_write(ssl, response, data, len(data))
         except Exception as error:
             logging.exception(error)
         finally:
@@ -306,8 +501,21 @@ def wsgi(ssl, response, info, user_data, aborted):
                 app_iter.close()
 
         if not headers_written:
-            write_headers(headers_set)
-        lib.uws_res_end_without_body(ssl, response, 0)
+            write_headers(headers_set)         
+        if is_chunked:
+            lib.uws_res_end_without_body(ssl, response, 0)
+        elif data_retry is not None:
+            _id = uuid.uuid4()
+            data_retry.id = _id
+            app._data_refs[_id] = data_retry
+            lib.uws_res_on_aborted(ssl, response, wsgi_on_data_ref_abort_handler, data_retry._ptr)
+            lib.uws_res_on_writable(ssl, response, wsgi_on_writable_handler, data_retry._ptr)
+        elif result is None or (not bool(result.has_responded) and bool(result.ok)): # not reachs Content-Length
+            logging.error(AssertionError("Content-Length do not match sended content"))
+            lib.uws_res_close(
+                ssl,
+                response
+            )
 
 
 def is_asgi(module):
@@ -332,7 +540,7 @@ class _WSGI:
         self.wsgi = app
         self.BASIC_ENVIRON = dict(os.environ)
         self.ws_compression = False
-
+        self._data_refs = {}
         self._ptr = ffi.new_handle(self)
         self.asgi_http_info = lib.socketify_add_asgi_http_handler(
             self.server.SSL, self.server.app, wsgi, self._ptr
